/*  
[INFO] 
NAME=Great Catalog Merger
AUTHOR=Vasyl Lukianenko
DEV=3DGROUND
DEV_WEB=https://3dground.net/
HELP=

[SCRIPT]
*/

global zipWrapper = undefined
global curlWrapper = undefined
global placerWarapper = undefined
global settingsWrapper = undefined
global AssetLibraryWrapper = undefined

struct zipWrapper (
    currentPath = getFileNamePath (getThisScriptFileName()),
	ftm = dotnetobject "System.Windows.Forms.Timer",
	updateTime = 0.5,
	queue = #(),
	onExtractedCallback = undefined,
	hingeLimit = 600,
	
	fn getStampDiff s = (
		return ((timeStamp() - s) / 1000.0) as integer
	),
		
	fn checkExtraction = (
		local d = #()
		
		for id in 1 to this.queue.count do (
			local q = this.queue[id]
			
			local sec = getStampDiff q.date
			if (sec > this.hingeLimit) do (
				append d id
				continue
			)
			
			if (doesFileExist q.archive) do continue
			
			q.extracted = true
			append d id
			
			if (this.onExtractedCallback != undefined) do try (this.onExtractedCallback q) catch ()
		)
		
		for i in d.count to 1 by -1 do try (deleteItem this.queue d[i]) catch()
		
		if (this.queue.count == 0) do this.ftm.Stop()
	),
	
	fn onTimerUpdate s a = (
		local self = s.tag.value
		
		try (
			self.checkExtraction()
		) catch (
			format "*** % ***\n" (getCurrentException())
			this.ftm.Stop()
		)
		
		(dotnetClass "system.gc").collect()	
		gc lights: true
	),
		
    fn extract q = (
		if (q == undefined) do return false
		
		local f = q.archive
		local o = q.dir
		
        if (f == undefined or not doesFileExist f) do return false
        if (getFilenameType f != ".zip") do return false
		
		this.ftm.Stop()
		this.ftm.Start()
		append this.queue q
       
		local args = f + ";" + o
		args = substituteString args "/" @"\"
		
		local cmd = "start /b /min extract.bat \"" +  args + "\""
		
		local curDir = sysInfo.currentDir
		sysInfo.currentDir = currentPath
		
		hiddenDosCommand cmd exitCode: &exitcode donotwait: true
		
		sysInfo.currentDir = curDir              
    ),
	
	fn dispose = (
		try (this.ftm.Stop()) catch ()
	),
	
	on create do (
		this.ftm.Interval = this.updateTime * 1000
		this.ftm.tag = dotNetMxsValue this
		dotNet.setLifetimeControl this.ftm #mxs
		dotnet.addEventHandler this.ftm "Tick" onTimerUpdate
	)
)

struct curlWrapper (
	currentPath = getFileNamePath (getThisScriptFileName()),
	queue = #(),
	done = #(),
	downloadDir = "",
	updateTime = 0.5,
	hingeLimit = 10,
	ftm = dotnetobject "System.Windows.Forms.Timer",
	onCompleteCallback = undefined,
	onProgressCallback = undefined,
	onHangCallback = undefined,
	ticks = 0,
		
	fn replaceSymbols s = (
		return substituteString s "/" @"\"
	),
	
	fn download url: "" name: "" id: "" cookies: "" = (
		this.ftm.Stop()
		this.ftm.Start()
			
		local archive = this.downloadDir + name + ".zip"
		local dump = this.downloadDir + name + ".dump"
		local dir = this.downloadDir + name + @"\"
		local date = timeStamp() as float
		
		if (doesFileExist dump) do return false
		
		struct tmp (url, dir, id, archive, dump, name, progress, size, extracted, date)
		append queue (tmp url: url dir: dir id: id archive: archive dump: dump name: name progress: 0 size: 0 extracted: false date: date)
		
		dump = this.replaceSymbols dump
		archive = this.replaceSymbols archive
		
		local args = dump + "|" + archive + "|" + cookies + "|" + url
		
		local cmd = "start /b /min download.bat \"" +  args + "\""
		
		local curDir = sysInfo.currentDir
		sysInfo.currentDir = currentPath
		
		HiddenDOSCommand cmd exitCode: &exitcode donotwait: true
		
		sysInfo.currentDir = curDir 
		
		return true
	),
	
	fn getStampDiff s = (
		return ((timeStamp() - s) / 1000.0) as integer
	),
	
	fn round i d: 1024.0 = (
		local k = ((i as float) / d) as integer
		return k as float
	),
	
	fn fileSize f = (
		return try ((getFileSize f) as float) catch (0.0)
	),
	
	fn findByName n = (
		local names = for i in this.queue collect i.name
		local id = findItem names n
		
		if (id == 0) do return undefined		
		return queue[id]
	),
	
	fn getSizeFromDump o = (
		if (not doesFileExist o) do return undefined
		local j = (o + ".r")
		copyFile o j
		
		local ms = MemStreamMgr.openFile j
		if (ms == undefined) do return undefined
		
		local s = 0
		
		while (not ms.eos()) do (
			local l = ms.readLine()
			if (findString l "Content-Length:" != undefined) do (
				local ss = (FilterString l ":")[2]
				if (ss == undefined) do continue
				ss = trimRight (trimLeft (ss))
				s = try (ss as integer) catch (0)
			)
		)
		
		memStreamMgr.close ms
		deleteFile j
		
		return try (s as float) catch (0.0)
	),
	
	-- Check by timer every 0.5s
	fn getDownloadProgress = (
		local d = #()
		local o = #()
		local hanged = #()
			
		for id in 1 to this.queue.count do (
			local q = this.queue[id]
			local sec = this.getStampDiff q.date
			
			-- If hanged add to list
			if (q.progress == 0.0 and sec > this.hingeLimit / 2.0) do (
				appendIfUnique hanged id
			)
			
			-- Clear hinged downloads
			if (q.progress == 0.0 and sec > this.hingeLimit) do (
				deleteFile q.archive
				append d id
				
				continue
			)
			
			if (q.size == 0.0) do (
				local total = this.getSizeFromDump q.dump
				if (total != undefined) then (
					q.size = total
				) else (
					continue
				)				
			)
			
			if (not doesFileExist q.archive) do continue
			
			local size = this.fileSize q.archive
			
			local p = (size / q.size * 100.0) as integer
			
			q.progress = p
			
			if (p == 100) do (
				-- Download complete
				append o q
				append d id					
			)
		)
		
		-- If really hanged and it's 70% - stop all
		local hangPercent = try (hanged.count / this.queue.count * 100.0) catch (0)
		if (hangPercent > 70.0 and this.onHangCallback != undefined) do (
			this.queue = #()
			this.onHangCallback hangPercent
			this.ftm.Stop()
		)
		
		if (this.onProgressCallback != undefined) do this.onProgressCallback this.queue
		
		for i in d.count to 1 by -1 do try (deleteItem this.queue d[i]) catch()
		
		for i in o do try (this.onCompleteCallback i) catch ()
		
		if (this.queue.count == 0) do this.ftm.Stop()
		
		return o
	),
		
	fn clearTempFiles = (
		for i in (getFiles (this.downloadDir + "*.dump")) do deleteFile i
	),
	
	fn isInstalled = (
		local f = @"c:\Windows\System32\curl.exe"
		return doesFileExist f
	),
	
	fn onTimerUpdate s a = (		
		local self = s.tag.value
		
		try (
			self.getDownloadProgress()
		) catch (
			format "*** % ***\n" (getCurrentException())
			self.ftm.Stop()
		)
		
		(dotnetClass "system.gc").collect()	
		gc lights: true
	),
	
	fn dispose = (
		try (this.ftm.Stop()) catch ()
	),
	
	on create do (		
		this.ftm.Interval = this.updateTime * 1000
		this.ftm.tag = dotNetMxsValue this
		dotNet.setLifetimeControl this.ftm #mxs
		dotnet.addEventHandler this.ftm "Tick" onTimerUpdate
	)
)

struct settingsWrapper (
    pth = getFilenamePath (getThisScriptFilename()),
    settingsFile = pth + @"settings.ini",
    fn s k v ss: "SETTINGS" = (
        setINISetting settingsFile ss k (v as string)
    ),
    fn g k type: #string  ss: "SETTINGS" = (
        local s = getINISetting settingsFile ss k
        case type of (
            #int: return s as integer
            #float: return s as float
            #bool: if s == "true" or s.count < 1 then true else false
            #bool_not: if s == "true" then true else false
            
            default: return s
        )
    ),
    fn a ss: "SETTINGS" type: #string = (
        local s = getINISetting settingsFile ss
        local o = #()
        
        for k in s do (
            local v = g k ss: ss type: type
            append o #(k, v)
        )
        
        return o
    ),
    fn d k ss: "SETTINGS" = (
        delIniSetting settingsFile ss k
    )
)

struct AssetLibraryWrapper (
    -----------------------------------------------------------------------------------------
    -- Globals
    -----------------------------------------------------------------------------------------
    dbg = false,
	currentPath = getFileNamePath (getThisScriptFileName()),
    cookies = "",
    url = "https://greatcatalog.net/3dmax-merger/",
	urlFallback = currentPath + "error.html",	
	urlErrorFirewall = currentPath + "error-firewall.html",	
    manifest = currentPath + "manifest.ini",
    ver = getINISetting manifest "INDEX" "VER",
    filePrefix = "great-catalog-",
    formTitle = "Great Catalog Merger " + this.ver + " (Under Construction)",
    placeModel = true,
    iconFile = currentPath + "icon.ico",
    wb = dotNetObject "System.Windows.Forms.WebBrowser",
    exec = dotNetObject "System.Windows.Forms.Textbox",
    u8 = dotNetObject "System.Text.UTF8Encoding",
    form = dotNetObject "MaxCustomControls.MaxForm",
	wndState = dotNetClass "System.Windows.Forms.FormWindowState",
    progressBarHeight = 4,    
    settings = settingsWrapper(),
    formDefaultSize = [800, 1100],
    formDefaultPos = [8, 26],
    formMinSize = [235, 300],
    bitmapTypes = #(),
	zip = zipWrapper(),
	curl = curlWrapper(),
	priorityFilesPtrn = "GC-*",
	outDir = @"", 
	   
    -----------------------------------------------------------------------------------------
    -- Events
    -----------------------------------------------------------------------------------------
    
	fn onExtractedSuccess q = (
		local cnt = this.upateTexturesForMaxFiles q.dir
		
		local cmd = "completeDownload"		
		if (cnt == 0) then (
			removeDir q.dir
			this.logging ("Can't extract archive: " +  q.id)
			cmd = "errorDownload"
		) else (					
			this.logging ("Extracted: " +  q.id)
		)
		
		this.runJavaScript "execCmd" #(#("cmd", cmd),  #("id", q.id))
	),
	
	fn onDownloadComplete q = (	
	
		if (not doesFileExist q.archive or getFileSize q.archive == 0) do (
			this.logging ("Wrong download (no file or zero size): " +  q.id)
			return false
		)
		
		this.logging ("! Download complete: " +  q.id)
		
		this.zip.extract q
	),
	
	fn onDownloadProgress qq = (
		for q in qq do (
			this.runJavaScript "execCmd" #(#("cmd", "updateProgress"),  #("id", q.id), #("value", q.progress))
		)
	),
	
	fn onHangDownloads p = (
		local u1= filenameFromPath (this.urlErrorFirewall)
		local u2= filenameFromPath (this.wb.url.toString())
		
		if (u1 == u2) do return false
		
		try (this.wb.Navigate this.urlErrorFirewall) catch ()
	),
	
    fn onResize s a = (
		local self = s.tag.value
		
        self.setControlsSize()
        self.saveWindowSize()
    ),
    
    fn onMove s a = (
		local self = s.tag.value
        self.saveWindowPos()
    ),
    
    fn onLoad s a = (
    ),
	
    fn onClose s a = (
		local self = s.tag.value
		
		self.zip.dispose()
		self.curl.dispose()
		self.curl = undefined
		self.zip = undefined
		
        self.saveWindowSize()
        self.saveWindowPos()
        self.saveWindowState()
      
        try (::__ASSETS_LIBRARY_VIEWER = undefined) catch ()
		
		-- BugFix: Random garbage collector error		
		(dotnetClass "system.gc").collect()	
		gc lights: true
    ),
       
    --- Handler
    fn onExecute s a = (
		local self = s.tag.value
		
        local t = s.text
        local c = (filterString t "=")[1]
        local h = (filterString t "#")
        local z1 = c.count + 2
        local z2 = -1
        try (z2 = t.count - z1 - h[h.count].count) catch ()
        
        local str = subString t z1 z2
        
        local values = filterString str "|"
        local v1 = values[1] as string
        local v2 = values[2] as string
        
        if (c == undefined) do return false
        
		self.logging ("Receive cmd: " +  c + " (" + str + ")")
		
        case c of (
            "DOWNLOAD": self.startDownload v1 v2
            "MERGE": self.startMerge v1
            "CHECK_EXIST_FILES": self.checkExistFiles values
            "SAVE_API_KEY": self.saveApiKey v1
            "SIGN_OUT": self.signOut()
            "OPEN_URL": self.openUrl v1
            "PASTE_TEXT": self.pasteText()
            "CHANGE_DOWNLOAD_FOLDER": self.changeDownloadFolder()
            "RELOAD": self.reload()
        )
		
		(dotnetClass "system.gc").collect()	
		gc lights: true
    ),
	
    fn onLoadCompleted s a = (
		local self = s.tag.value
		local c = try (self.wb.Document.Cookie) catch ("")		
        self.cookies =self.getSession c
		
		(dotnetClass "system.gc").collect()	
		gc lights: true
    ),
	
    on create do (
        this.init()
    ),
    
    -----------------------------------------------------------------------------------------
    -- Form
    -----------------------------------------------------------------------------------------
    
	fn getRealMaxVersion = (
		local i = dotNetClass "System.Diagnostics.FileVersionInfo"
		local fv = i.GetVersionInfo ((getDir #maxroot) + "3dsmax.exe")
		return filterString fv.FileVersion "."
	),

	
	fn setInViewForm = (
		local x = this.form.Location.x
		local y = this.form.Location.y
		local w = this.form.Width
		local h = this.form.Height
		
		local s = sysInfo.desktopSize
		local maxX = abs(s.x - w) + 6
		local maxY = abs(s.y - h) - 40
		
		if (x > maxX) do this.form.Location.x = maxX
		if (y > maxY) do this.form.Location.y = maxY
		if (x < 0) do this.form.Location.x = 0
		if (y < 0) do this.form.Location.y = 0
	),
	
    public fn showForm = (
        this.setWindowPos()
        this.setWindowSize()
        this.setWindowState()
		
        this.form.Text = this.formTitle
        this.form.ShowInTaskbar = false
        try (this.form.Icon = dotNetObject "System.Drawing.Icon" this.iconFile) catch ()
        
        try (this.form.WindowState = this.wndState.Normal) catch ()
        try (this.form.StartPosition = this.form.StartPosition.Manual) catch ()
        
        --this.form.Show()
        --this.form.topmost = true
        
        try (this.form.ShowModeless()) catch ()
        
        this.form.Focus()
		this.setInViewForm()	
    ),
    
    fn closeForm = (
        try (this.form.Close()) catch ()
    ),
    
    fn getWidnowState = (
		try (
			case this.form.WindowState of (
				(this.wndState.Maximized): return #maximized
				(this.wndState.Minimized): return #minimized
				default: return #normal
			)
		) catch ()
    ),
    
    fn setControlsSize = (
        wb.Location = dotNetObject "System.Drawing.Point" 0 0
        wb.Size = this.form.ClientSize      
    ),
    
    fn setWindowPos = (
        local x = this.settings.g "POSX" type: #int
        local y = this.settings.g "POSY" type: #int
        local p = this.formDefaultPos
        
        if (x == 0) do x = p.x
        if (y == 0) do y = p.y
        
        this.form.Location = dotNetObject "System.Drawing.Point" x y
    ),
    
    fn setWindowSize = (
        local w = this.settings.g "WIDTH" type: #int
        local h = this.settings.g "HEIGHT" type: #int
        local s = this.formDefaultSize
        local m = this.formMinSize
        
        if (w == 0) do w = s.x
        if (h == 0) do h = s.y
        
        this.form.Size = dotNetObject "System.Drawing.Size" w h
        this.form.MinimumSize = dotNetObject "System.Drawing.Size" m.x m.y
    ),
    
    fn setWindowState = (
        local v = this.settings.g "STATE"
        
		case v of (
            "maximized": this.form.WindowState = this.wndState.Maximized
            default: this.form.WindowState = this.wndState.Normal
        )
    ),
    
    --------------------
    
    fn saveWindowState = (
        local s = this.getWidnowState()
        this.settings.s "STATE" s
    ),
    
    fn saveWindowPos = (
        local s = this.getWidnowState()
        
        if (s != #maximized) do (
            local x = this.form.left
            local y = this.form.top
            
            if (x < -31000 or y < -31000) do return false
            
            this.settings.s "POSX" x
            this.settings.s "POSY" y
        ) 
    ),
    
    fn saveWindowSize =(
        local s = this.getWidnowState()
        
        if (s == #normal) do (
            local w = this.form.size.width
            local h = this.form.size.height
            
            this.settings.s "WIDTH" w
            this.settings.s "HEIGHT" h
        )
    ),
    
    -----------------------------------------------------------------------------------------
    -- Func
    -----------------------------------------------------------------------------------------
    
	fn loadFileAsArray o = (
		if (not doesFileExist o) do return #()
		
		local f = openFile o
		local b = #()

		if (f == undefined) do return #()
		
		while (not eof f) do (
			l = readline f				
			append b l
		)
		
		flush f
		close f
		
		return b
	),
	
	fn md5 t = (
		local dnMD5Provider = dotNetObject "System.Security.Cryptography.MD5CryptoServiceProvider"
		local dnASCIIEncoding = dotNetObject "System.Text.ASCIIEncoding"
		local gb = dnASCIIEncoding.GetBytes t

		local hashBytes = dnMD5Provider.ComputeHash gb
		local dnStringBuilder = dotnetObject "System.Text.StringBuilder"
		local dnString = dotNetClass "System.String"
		for h in hashBytes do (
			dnStringBuilder.Append (dnString.Format "{0:X2}" h)
		)	
		
		return dnStringBuilder.ToString()
	),
	
	fn getUniqueId = (
		-- Get Disk Drive ID
		local p = getFilenamePath(getThisScriptFilename())
		local f = "uid.data"
		local ff = p + f
		HiddenDOSCommand ("wmic DISKDRIVE get SerialNumber > " + f) startpath: p
		local aa = loadFileAsArray ff
		deleteFile ff
		for i in aa do (
			if (findString i "SerialNumber" != undefined) do continue
			i = trimLeft (trimRight i)
			if (i != undefined and i.count > 5) do return md5(i)
		)
		
		-- Get registyr GUID if not found DRIVE ID
		local keyPath = @"SOFTWARE\Microsoft\Cryptography"
		registry.openKey HKEY_LOCAL_MACHINE keyPath accessRights: #readOnly key: &key
		registry.queryValue key "MachineGuid" type: &type value: &val
		registry.flushKey key
		registry.closeKey key
		
		if (val != undefined and val.count > 8) do return md5(val)
		
		-- If nothing found return hash from compname and username
		return md5(sysInfo.computername + sysInfo.username)	
	),
	
	fn setUniqueId = (
		local uid = settings.g "UID"
		if (uid.count < 5) do (
			local id = this.getUniqueId()
			settings.s "UID" id
		)
	),
	
	fn getSession c = (
		if (c == undefined) do return ""
		
		for i in (filterString c ";") do (
			if (findString i "SESSID" != undefined) do return trimLeft (trimRight i)
		)
		
		return ""
	),
		
	fn checkInternet = (
		return internet.CheckConnection url: this.url force: true
	),
	
    fn logging s = (
        if (not this.dbg) do return false
        print ("[" + localTime + "] >>> " + s as string)
    ),
    
    fn boost = (
        if (((maxVersion())[1] / 1000) >= 13) do windows.processPostedMessages() 
    ),
    
    fn pasteText = (
        local t = getclipboardText()
        
        -- SHA256 check
        if (t == undefined or t.count != 64) do return undefined
        
        this.runJavaScript "execCmd" #(#("cmd", "pasteText"),  #("text", t))
    ),
       
    fn getFileFromUrl s = (
        p = findString s "?"
        if (p != undefined) do (
            s = subString s 1 (p - 1)
        )
        
        s = filenameFromPath  s
            
        return s
    ),
    
    fn getUri s = (
        s += "?" + timestamp() as string
        return dotNetObject "System.Uri" s
    ),
    
    fn createJSON data quotes: true newLine: true = (
        local q = ""
        local n = ""
        
        if (quotes) do q = "\""
        if (newLine) do n = "\n"
        
        str = "{"
        for i in 1 to data.count do (
            str += "\"" + (data[i][1] as string) + "\": "
            str += q + (data[i][2] as string) + q
            if (i < data.count) do str += "," + n
        )
        str += "}"
        
        return str
    ),
    
    fn updateDownloadFolder = (
        local downloadFolder = settings.g "DOWNLOAD_FOLDER"
        
        if (downloadFolder == "" or not doesDirectoryExist downloadFolder) do return false
        
        this.outDir = substituteString downloadFolder @"\" "/"
		this.curl.downloadDir = this.outDir
    ),
    
    fn changeDownloadFolder = (
        local p = getSavePath title: "Choose download folder"
        if (p == undefined) do return false
        p += @"\"
        
        settings.s "DOWNLOAD_FOLDER" p
        
        this.updateDownloadFolder()
        
        local u = this.url + "settings/"
        
        this.navigate u reload: true
    ),
	
	fn reload = (
		this.navigate this.url reload: true
	),
    
    fn navigate s reload: false = (
		if (not this.checkInternet()) do (
			try (this.wb.Navigate this.urlFallback) catch ()
			return false
		)
		
        local data = #()
        local apiKey = settings.g "API_KEY"
		local uid = settings.g "UID"
        
        header = "Content-Type: application/json\r\n"
        header += "Browser: mxs\r\n"
        
        this.updateDownloadFolder()
        
        this.placeModel = settings.g "PLACE_MODEL" type: #bool
        
        append data #("computername", sysInfo.computername)
        append data #("username", sysInfo.username)
        append data #("uid", uid)
        append data #("api-key", apiKey)
        append data #("download-folder", this.outDir)
        append data #("place-model", this.placeModel)
        append data #("version", this.ver)
        
        local bytes = this.u8.GetBytes (this.createJSON data)
        
        if (reload) do (
            local a = if (this.find s "?") then "&" else "?"
            s += a + "time=" + (timeStamp() as string)
        )
        
        try (this.wb.Navigate s "" bytes header) catch ()	
    ),
    
    fn runJavaScript s data = (
        str = "{"
        for i in 1 to data.count do (
            str += "\"" + (data[i][1] as string) + "\":\"" + (data[i][2] as string) + "\""
            if (i < data.count) do str += ", "
        )
        str += "}"
        
		this.logging ("Run cmd: " +  str)
		
        try (wb.Document.InvokeScript s #(str)) catch (
            format "*** % ***\n" (getCurrentException())
        )
    ),
    
    fn checkExistFiles v = (
      
        if (v == undefined) do return false
        
        local o = ""
        
        for id in v where (id != undefined and id != "") do (
            id = getFilenameFile id
            
            local d = this.outDir + this.filePrefix + id + @"\"
            
            if (doesDirectoryExist d) do (
                o += id + "|"
            )
        )
        
        this.runJavaScript "execCmd" #(#("cmd", "existItems"),  #("ids", o))
    ),
    
    fn saveApiKey v = (
        if (v == undefined or v == "" or v.count < 8) do return false
        
        settings.s "API_KEY" v
    ),
    
    fn signOut = (
        settings.d"API_KEY"
        this.navigate (url + "?api_signout=1&api=3dmaxmerger")
    ),
    
    fn openUrl v = (
        if (v == undefined) do return false
        v = v as string
        
        try (
            ShellLaunch v ""
        ) catch (
            this.logging ("Can't redirect to URL: " + v)
        )
    ),
    
    fn init = (
		local ver = getRealMaxVersion()
		-- 2019 3Ds Max without SP, (Bug with Dotnet Events)
		if (ver[1] == "21" and ver[4] == "845") do (
			messageBox "Unfortunately the script does not work on version 3Ds Max 2019!\nPlease install the latest Service Pack!" title: "Error!"
			return #abort
		)
		
		this.setUniqueId()
		
		try (append bitmapTypes #(Bitmaptexture, #filename)) catch ()
        try (append bitmapTypes #(CoronaBitmap, #filename)) catch ()
        try (append bitmapTypes #(VRayBitmap, #HDRIMapName)) catch ()
        try (append bitmapTypes#(VRayHDRI, #HDRIMapName)) catch ()
		
        dotNet.addEventHandler this.form "Resize" onResize
        dotNet.addEventHandler this.form "Move" onMove
        dotNet.addEventHandler this.form "Load" onLoad
        dotNet.addEventHandler this.form "Closing" onClose
        dotNet.addEventHandler this.exec "TextChanged" onExecute
        dotnet.addEventHandler this.wb "DocumentCompleted" onLoadCompleted
		this.wb.tag = dotNetMxsValue this
		this.exec.tag = this.wb.tag
		this.form.tag = this.wb.tag
		
        this.exec.Size = dotNetObject "System.Drawing.Size" 0 0
        this.exec.Location = dotNetObject "System.Drawing.Point" -100 -100
        
        this.wb.ObjectForScripting = exec
        this.wb.ScriptErrorsSuppressed = not this.dbg
        try (this.wb.AllowDrop = false) catch ()
        this.wb.AllowWebBrowserDrop = false
		
		local lifeTime = #mxs
		
		dotNet.setLifetimeControl this.exec lifeTime
		dotNet.setLifetimeControl this.wb lifeTime
		dotNet.setLifetimeControl this.form lifeTime
		
        this.form.controls.add  this.exec
        this.form.controls.add  this.wb
       	
        this.setControlsSize()
      
        this.showForm()
		
        makeDir this.outDir all: true
		
		-- Setup Zip and Curl
		this.curl.downloadDir = this.outDir
		this.curl.clearTempFiles()
		this.curl.onCompleteCallback = this.onDownloadComplete
		this.curl.onProgressCallback = this.onDownloadProgress
		this.curl.onHangCallback = this.onHangDownloads
		this.zip.onExtractedCallback = this.onExtractedSuccess
		
		this.navigate this.url
		
        return #form_created
    ),
    
    fn getCurrentMaxVerson = (
        local v = maxVersion()
        v = (v[1] / 1000) - 2
        return 2000 + v
    ),
    
    fn isCanOpenMaxFile f = (
        local fv = try ((getMaxFileVersionData f)[1]) catch (undefined)
        local mv = (maxVersion())[1]
        
        if (fv == undefined) do return false
        
        return fv <= mv
    ),
    
    fn formatToVersion v = (
        local vv = substituteString v "3Ds Max " ""
        vv = trimLeft (trimRight vv)
        return try (vv as integer) catch #unknown
    ),
	
	fn sortFilesByPriority srt = (
		fn compareCallback v1 v2 = (
			local f1 = getFilenameFile v1
			local f2 = getFilenameFile v2
			local ptrn = this.priorityFilesPtrn
			if (not MatchPattern f1 pattern: ptrn and MatchPattern f2 pattern: ptrn) do (
				return 1
			)
			
			return 0
		)
		
		qsort srt compareCallback
		
		return srt
	),
    
    fn getUnzippedMaxFile o = (
        ff = for i in (this.getFilesRecursive o ptrn: "*.max") where this.isCanOpenMaxFile i collect i
        if (ff.count == 0) do return undefined
        
		ff = this.sortFilesByPriority ff
		
        local r = this.getRender()
        if (r == undefined) do return ff[1]
        
        for i in ff do (
            n = toLower (getFilenameFile i)
            for ii in r do (
                if (findString n ii != undefined) do return i
            )
        )
        
        return ff[1]
    ),
    
    fn getRender = (
        r = renderers.current as string
        
        if matchpattern r pattern:"*Corona*" do return #("corona")
        if matchpattern r pattern:"*V_Ray*" do return #("vray", "v-ray", "v_ray")
        if matchpattern r pattern:"*Default_Scanline*" do return #("scanline")
        if matchpattern r pattern:"*mental_ray*" do return #("mentalray")
        if matchpattern r pattern:"*iray_Renderer*" do return #("iray")
        if matchpattern r pattern:"*maxwell*" do return #("maxwell")
        if matchpattern r pattern:"*Arnold*" do return #("arnold")
            
        return undefined
    ),
    
     
    fn isValidBitmap b prop = (	
        if(not hasProperty b prop) do return false
        local p = getProperty b prop
        return b != undefined and p != undefined and p != "" and p.count != 0
    ),
    
    fn getBitmapFile b prop = (
        if(not hasProperty b prop) do return undefined
        return getProperty b prop
    ),

    fn setBitmapFile b prop v: undefined = (
        if(not hasProperty b prop) do return false
        setProperty b prop v
        return true
    ),
    
    fn getFilesRecursive root ptrn: "*" = (
        local dd = getDirectories (root + "/*")
        local ff = getFiles (root + "/" + ptrn)
        for i in dd do join ff (getFilesRecursive i ptrn: ptrn)
        
        return ff
    ),
    
	fn updateTexturesPaths f = (
		local p = getFilenamePath f
		local d = getMAXFileAssetMetadata f
		
		local files = this.getFilesRecursive p
		local names = for i in files collect filenameFromPath i
		
		for i in d where i != undefined and i.filename != undefined and i.filename != "" do (		
			local n = filenameFromPath i.filename
			
			local id = findItem names n
			if (id == 0) do continue
			
			i.filename = i.resolvedFilename = files[id]		
		)
		
		setMAXFileAssetMetadata f d
		
		d = undefined
		files = undefined
		names = undefined
	),
	
	fn upateTexturesForMaxFiles root = (
		local t = 0
		for f in (this.getFilesRecursive root ptrn: "*.max") do (
			this.updateTexturesPaths f
			t += 1
		)
		
		return t
	),
	
    fn setTexturesPath path: "" geom: selection = (
        local files = getFilesRecursive path
        local names = for i in files collect filenameFromPath i
              
        for bt in bitmapTypes where (bt[1] != undefined) do (
            local type = bt[1]
            local prop = bt[2]
            
            for g in (geom as array) do (
                for b in getClassInstances type target: g where (isValidBitmap b prop) do (
                    local f = getBitmapFile b prop
                    local n = filenameFromPath f
                    
                    if (findString f path != undefined) do continue
                    
                    local id = findItem names n
                    if (id == 0) do continue
                    
                    setBitmapFile b prop v: files[id]
                )
            )
        )
    ),
	
	fn recursiveShowMaps om t = (
		local c = getNumSubMtls om
		if (c >= 1) then (
			for i in 1 to c do (
				sm = getSubMtl om i
				if sm != undefined do (
					recursiveShowMaps sm t
				)
			)
		) else(
			try(showTextureMap om t)catch()
		)
	),

	-- Display maps in viewport

	fn toggleVpMaps t scope: selection = (
		for o in scope do(
			if o.material != undefined do(
				recursiveShowMaps o.material t
			)
		)
	),
    
    fn refreshPivotPos = (
        MaxOps.pivotMode = #pivotOnly
        ForceCompleteRedraw()
        max scale
        ForceCompleteRedraw()
        max rotate
        ForceCompleteRedraw()
        max move
        ForceCompleteRedraw()
        MaxOps.pivotMode = #none
        ForceCompleteRedraw()
    ),
    
    fn downloadModelIni id d = (
        local u = this.url + id + ".ini"
        return this.downloadFile u (d + "index.ini")
    ),
    
	fn groupObjects s n = (
		local g = for i in s where i.parent == undefined collect i
		case of (
			(g.count == 1): return g[1]
			(g.count > 1): return group g name: n
		)
		
		return group s name: n
	),
	
	fn showStatus t c: true = (
		global loadingIndicator_StatusText = t
		global loadingIndicator_timerActive = c
		global loadingIndicator
		
		try(destroyDialog loadingIndicator) catch()
		
		rollout loadingIndicator "Loading..." (
			label lblStatus "Loading..." offset: [0, 14] 
			timer tmrTimer "" active: true interval: 500
			
			on loadingIndicator open do (
				lblStatus.caption = loadingIndicator_StatusText
				tmrTimer.active = loadingIndicator_timerActive
			)
			
			on tmrTimer tick do 
			(	
				try(destroyDialog loadingIndicator)catch()
				tmrTimer.active = false
			)
		)
		createDialog loadingIndicator 160 50 style: #(#style_border)
	),
	
    fn startMerge id = (		
        local id = getFileNameFile id
        local d = this.outDir
        
        local f = d + id + ".zip"
        local o = d + this.filePrefix + id + @"\"
        
        local mm = this.getUnzippedMaxFile o
        
        if (mm == undefined) do (
            m = "Can't find correct *.max file in: " + o + " !"
            m += "\n\n"
            m += "Possible reasons:"
            m += "\n"
            m += "1. Archive not downloaded\n"
            m += "2. Archive not unzipped\n"
            m += "3. 3Ds Max file of the required version was not found in the folder\n"
            m += "4. The folder does not contain a 3Ds Max file for your Render"
            
            messageBox m title: "Error!"
            return false
        )
        
        if (not this.isCanOpenMaxFile mm) do (
            m = "Unfortunately, the version of the 3D Model is higher than your version of the 3Ds Max. Therefore, Merge is not possible.\n\nDo you want to open the folder with the model and try it manually using FBX?!"
            q = queryBox m title: "Confirm!"
            if (q) do shellLaunch o ""
            return false
        )
        
        this.logging ("Merge file: " + mm)
		
		this.runJavaScript "execCmd" #(#("cmd", "startMerge"),  #("id", id))
		showStatus "Wait while merging..."
        
        --mergeMaxFile mm #select #promptDups  #promptMtlDups  quiet:true
        mergeMaxFile mm #select #autoRenameDups #useSceneMtlDups  quiet:true
		
        sel = selection as array
        
        this.refreshPivotPos()
		this.toggleVpMaps true
		
		-- Group on the fly selected objects
		local g = this.groupObjects sel id
		try (select g) catch ()
		
        -- Temporary disabled need to set tex before merge
		--this.setTexturesPath path: o geom: sel
        
        toolMode.selectioncenter()
		
        /*max tool zoomextents all
		max zoom out 2x all
		max zoom out 2x all*/
		
		this.runJavaScript "execCmd" #(#("cmd", "completeMerge"),  #("id", id), #("count", sel.count as string))	
        
        this.placeMode()
        
        return true
    ),
    
    fn find s f = (
        return findString s f != undefined
    ),
    
    fn startDownload u f = (		
		if (not this.curl.isInstalled()) do (
			local m = "This script only supports the latest builds of Windows 10!\nPlease update your operating system in order to use " + this.formTitle
			return messageBox m title: "Error!"
		)
		
		local id = getFilenameFile f
		local n = this.filePrefix + id
		
		this.curl.download url: u id: id name: n cookies: this.cookies
    ),
    
    fn downloadFile u dl = (
        deleteFile dl
        
        local w = dotNetObject "System.Net.WebClient"
        local p = if (this.find u "?") then "&" else "?"
        
        try (
            local spm = dotNetClass "System.Net.ServicePointManager"
            spm.SecurityProtocol = spm.SecurityProtocol.TLS12
            
            h = dotnetclass "System.Net.HttpRequestHeader"
            w.Headers.set (h.cookie) this.cookies
            
            w.DownloadFile (u + p +"time=" + timestamp() as string) dl
            try (w.Dispose()) catch ()
        ) catch (
            return false
        )
        
        return true
    ),
    
    fn placeMode = (		
		try (fileIn (currentPath + "QuickPlacer.ms")) catch()
    )
)

struct checkerWrapper (
	settings = settingsWrapper(),
	fn setDownloadDirDialog = (
		global isGCMergerDownloadFolderSetup = false
		try (DestroyDialog ::rSetDownloadDir) catch ()
		rollout rSetDownloadDir "GC Merger - Download Folder Setup" (
			local settings = undefined
			
			group "Information" (
				label lblText1 "Please select a folder to download the models!" align: #left
				label lblText2 "Gigabytes of data will be stored in this folder" align: #left
				label lblText3 "make sure you have enough hard drive space." align: #left
			)
			
			group "Setup Download Folder" (				
				edittext edtOutDir "" readOnly: true height: 26 width: 245 across: 2
				button btnSelectOutDir "..." width: 40 height: 26 align: #right
				button btnApply "Apply" height: 30 width: 120 offset: [0, 5]
			)
			
			HyperLink hpSite "What is GC Merger?" address: "https://greatcatalog.net/en/3dmax-merger/" color: orange align: #center hoverColor: gray visitedColor: orange
			
			on rSetDownloadDir open do (
				settings = this.settings
			)
			
			on btnSelectOutDir pressed do (
				local p = getSavePath title: "Choose download folder"
				if (p == undefined) do return false
				p += @"\"
				
				edtOutDir.text = p
			)
			
			on btnApply pressed do (
				if (edtOutDir.text != "" and isDirectoryWriteable (trimRight edtOutDir.text @"\")) do (
					settings.s "DOWNLOAD_FOLDER" edtOutDir.text
					::isGCMergerDownloadFolderSetup = true
					try (DestroyDialog ::rSetDownloadDir) catch ()
					return true
				)
				
				messageBox "Files are not allowed to be written to this folder.\nPlease choose another folder!" title: "Error!"
				
				return false
			)
		)
		CreateDialog ::rSetDownloadDir 320 200 modal: true
		
		return ::isGCMergerDownloadFolderSetup
	),
	fn checkDownloadDir = (
		local outDir = settings.g "DOWNLOAD_FOLDER"
		if (outDir != "" and doesDirectoryExist outDir and isDirectoryWriteable (trimRight outDir @"\")) then (
			return true
		)
		
		return setDownloadDirDialog()
	)
)

(
	local isDownloadDirExist = (checkerWrapper()).checkDownloadDir()
	
	if (isDownloadDirExist) do (
		try (
			::__ASSETS_LIBRARY_VIEWER.showForm()
		) catch (
			gc lights: true
			(dotnetClass "system.gc").collect()
			global __ASSETS_LIBRARY_VIEWER = AssetLibraryWrapper()
		)
	)
)