/*  
[INFO] 
NAME=Great Catalog Merger Installer
AUTHOR=Vasyl Lukianenko
DEV=3DGROUND
DEV_WEB=https://3dground.net/
HELP=

[SCRIPT]
*/

struct installGCMerger (
    currentPath = getFileNamePath (getThisScriptFileName()),
    manifest = currentPath + "manifest.ini",
        
    fn settings k s: "INDEX" = (
        return getINISetting this.manifest s k
    ),
    
    macro = this.settings "MACRO",
    cat = this.settings "CAT",
    txt = this.settings "TXT",
    run = this.settings "RUN",
    ver = this.settings "VER",
    embedPath = substituteString this.currentPath @"\" "\\\\",
    upd = this.embedPath + "update.ms",
	logStruct = this.embedPath + "log.ms",
    script = this.embedPath + this.run,
	iconFile = "GC",

    fn installMacroScript = (
        local exec = ""
        exec += "macroScript " + this.macro + "\n"
        exec += "buttontext: \"" + this.txt + "\"\n"
        exec += "category: \"" + this.cat + "\"\n"
        exec += "icon: #(\"" + this.iconFile + "\", 1)\n"
        exec += "(\n"
        exec += "\ton execute do (\n"
        exec += "\t\tlocal s = \"" + this.script + "\"\n"
        exec += "\t\tlocal u = \"" + this.upd + "\"\n"
        exec += "\t\ttry(fileIn(s)) catch(messageBox \"Script not found! Download " + this.txt + " again!\" title: \"Error!\")\n"
        exec += "\t\ttry(fileIn(u)) catch()\n"
        exec += "\t)\n"
        exec += ")\n"
        
        try (execute exec) catch (print "Can't install MacroScript")
    ),
    
    fn addQuadMenuButton remove: false = (
        local quadMenu = menuMan.getViewportRightClickMenu #nonePressed
        local theMenu = quadMenu.getMenu 1
        
        fn findMenuItem theMenu menuName = (
            for i in 1 to theMenu.numItems() where (theMenu.getItem i).getTitle() == menuName do return i
            return 0
        )
        
        fn unregisterMenuItem theMenu menuName = (
            try (
                for i in 1 to theMenu.numItems() do (
                    if((theMenu.getItem i).getTitle() == menuName) do (
                        theMenu.removeItemByPosition i 
                        
                        if((theMenu.getItem (i - 1)).getIsSeparator()) do theMenu.removeItemByPosition (i - 1)
                    )
                )
            ) catch()
        )
        
        local item = try(findMenuItem theMenu "Select &Similar") catch (6)
        
        unregisterMenuItem theMenu this.txt
        
        if(not remove) do (
            quadItem = menuMan.createActionItem this.macro (this.cat)
            theMenu.addItem quadItem (item += 1)
        )
            
        menuMan.updateMenuBar()
    ),
    
    fn getNotes = (
        return getINISetting this.manifest this.ver
    ),
	
	fn copyIcons = (
		local p1 = currentPath + @"icons\"
		local p2 = (getDir #userIcons) + @"\"
		
		for f in getFiles (p1 + "*") do (
			local n = filenameFromPath f
			copyFile f (p2 + n)
		)
	),

    on create do (
        local n = "\r\n"
        
		this.copyIcons()
        this.installMacroScript()
        
        addQuadMenuButton remove: true
        addQuadMenuButton remove: false
        
		colorman.reInitIcons()
		
        deleteFile (this.currentPath + "mzp.run")
        
        local m = n
        m += "\t==============================================" + n
        m += "\t=====     " + this.txt + " " + this.ver + " installed successfully!" + "      =====" + n
        m += "\t==============================================" + n + n + n
        
        m += this.txt + " added automatically to Quad Menu!" + n + n + n
        m += "[!] To add a button to the toolbar:" + n
        m += " 1. Go to Customize User Interface" + n
        m += " 2. Tab: Toolbars" + n 
        m += " 3. Category: \"" + this.cat + "\"" + n
        m += " 4. Drag&Drop the \"" +  this.txt + "\" to toolbar" + n + n + n
        
        m += "[?] Changelog:" + n
        for i in getNotes() do m += " " + i + n
		try (
			execute ("fileIn \"" + this.logStruct + "\"")
			(logGCMergerWrapper()).show m title: "Installation"
		) catch ()		
        try (execute ("fileIn \"" + this.script + "\"")) catch ()
    )
)

installGCMerger()